% CompetitiveAttractorBPNetDemo.m 
% Competitive network demonstration program to add in an attractor functionality and also backprojection functionality
% 20 December 2020
% This program calls functions NormVeclen.m and spars.m
%
% Please do not redistribute this code or modifications of it without the following notice:
% (c) E T Rolls www.oxcns.org 2021
% The background to these networks is provided in
% Rolls ET (2021) Brain Computations: What and How (Oxford University Press) Appendix 2 (B)
% The instructions and exercises for this code are in
% NeuronalNetworkSimulationSoftware.pdf which is 
% Rolls ET (2021) Brain Computations: What and How (Oxford University Press) Appendix 4 (D)
% which is available at https://www.oxcns.org by following the link to the book.
% This code is available at https://www.oxcns.org/publications/html in
% CortexCompAttBP20.zip as Supplementary Material to 
% Rolls,E.T. (2021) The connections of cortical pyramidal cells can implement
% categorisation, attractor memory, and top-down recall and attention. Brain Sructure and Function

% This Matlab program is intended to demonstrate some of the principles of how competitive and related networks operate,
% and for that reason Matlab code optimizations are frequently eschewed in favor of illustrating principles of neuronal operations 

% The synaptic weight matrix (SynMat) has column vectors to represent the synapses on a single neuronn.
% There is one column vector therefore for each of the N neurons.
% Thus SynMat(Row, Column) implements SynMat(synapse, neuron)

% The program will pause between steps, and wait for a keyboard 'Enter'
% The synaptic matrix has 3 parts:
% nSyn   1 : 100 is for the backprojections (with nSyn=100)
% nSyn 101 : 200 is for the recurrent collaterals for the attractor network
% nSyn 201 : 300 is for the forward inputs

clear all; close all hidden; format compact; % format bank;
fig = 0;
% rng(1);         % set random number generator only for code development
N = 100;          % number of neurons in the competitive part of the net
nSyn = N;         % number of synapses on each neuron that are not deleted. Set the dilution by the divisor
nepochs = 10;     % number of training epochs. default 20
NRecEpochs = 10;   % number of short-term memory recall epochs
Learnrate = 0.1;  % 0.1 works and is the default. Aim for weightchange = 0.02 - 0.2. For the competitive learning
LearnrateRC = 0.03; % for the recurrent collateral learning for the attractor. Default 0.03. Works in range 0.01 to 0.1
LearnrateBP = 0.1;% for the Back Projection inputs. Default 0.1.
if Learnrate == 0
    nepochs = 0; 
end
OutputSparseness = 1 / N; % the sparseness of the output firing. Default 0.01 for N = 100
display = 1; % 0 no display; 1 display during learning epoch 1
Actvn = zeros(1, N);  % activation
Rate = zeros(1, N);   % output firing rates
ScaleFactAttToComp = 0.1; % ratio of attractor recurrent input to external competitive input.  Default = 0.1
ScaleFactBP = 0.1;    % ratio of backprojection input to external competitive input. Default = 0.1

nPatts = 30; % default 30, especially for graded
TrainPatts = zeros(nSyn, nPatts);
PattType = 1; % 1 for graded overlap, else random
if PattType == 1
    % generate training patterns overlapping by shift 
    leng = 20;       % length of each training pattern
    InputSparseness = leng / nSyn;  
    shift = 3;      % shift between training patterns 
    for patt = 1 : nPatts
        for x = 0 : leng -1
            syn = (patt -1) * shift + x;
            syn = mod(syn, nSyn);
            TrainPatts(syn + 1 , patt) = 1;
        end    
        if (patt - 1) * shift + leng >= nSyn
            break
        end
    end
    nPatts = patt; % this will set nPatts to be the number that can be fitted into the array without much wrap-around
    disp(['Number of patterns generated= ', num2str(nPatts), ' with shift of ', num2str(shift), ' and length of ', num2str(leng)]);
else
    % or generate some random binary patterns
    for patt = 1 : nPatts
        randpatt = zeros(nSyn, 1);
        r = randperm(nSyn);
        r = r(1 : nSyn/10); % sparseness = 0.1 (e.g. get 10 random numbers in the range 1 : 100)
        randpatt(r) = 1;
        TrainPatts(:, patt) = randpatt;
    end
end
for patt = 1 : nPatts
    TrainPatts(:, patt) = TrainPatts(:, patt) / norm(TrainPatts(:, patt));     % normalize the length of each training pattern    
end

if display == 1
    fig = fig + 1;
    figure(fig)
    imagesc(TrainPatts);
    colormap(gray);
    title(['Training patterns']); xlabel('Pattern number'); ylabel('Forward input neuron number');
    %pause
end

% The BP recall pattern will be linearly related to the training pattern number. Set this up next
BPpatts = zeros(nSyn, nPatts); % the recall patterns
leng = floor(nSyn / nPatts); % length of each pattern
for patt = 1 : nPatts
    BPpatt = zeros(nSyn, 1);
    start = (patt-1) * leng + 1;
    BPpatt(start : start + leng -1) = 1;
    BPpatts(:, patt) = BPpatt;
end
% for patt = 1 : nPatts
%     BPpatts(:, patt) = BPpatts(:, patt) / norm(BPpatts(:, patt));     % normalize the length of each training pattern    
% end
if display == 1
    fig = fig + 1;
    figure(fig)
    imagesc(BPpatts);
    colormap(gray);
    title(['Backprojection patterns']); xlabel('Pattern number'); ylabel('Backprojection input neuron number');
end

SynMat = rand(nSyn * 3, N); % (Synapses = Rows, neurons = Columns). The first nSyn synapses (at the top of the Matlab figure) are for the backprojection input, 
% the second nSyn are for the recurrent collaterals for the attractor; the third nSyn are for the forward inputs (closest to the cell body
for neuron = 1 : N
    SynMat(:, neuron) = NormVeclen(SynMat(:, neuron));  % normalize the synatic weight vector on each neuron
end

% Train the synaptic matrix for the compet forward input and the Recurrent Collaterals and the BackProjions SIMULTANEOUSLY  
MeanPreSynapticRate = sum(TrainPatts(:,1)) / nSyn;
for epoch = 1 : nepochs
    order = randperm(nPatts); % present the patterns in random order during training
    for i = 1 : nPatts
        patt = order(i);
        % calculate activation
        for neuron = 1 : N
            PreSynInput = BPpatts(:, patt) * ScaleFactBP; % the BP input for pattern association learning
            PreSynInput = [PreSynInput; Rate(:) * ScaleFactAttToComp]; % add the recurrent collateral input
            PreSynInput = [PreSynInput; TrainPatts(:, patt)]; % the forward input for competitive learning
            Actvn(neuron) = dot(PreSynInput, SynMat(:, neuron)); % dot product of input pattern and synaptic weight vector for one neuron
        end
        scale = 1.0 / (max(Actvn) - min(Actvn));
        Actvn = (Actvn - min(Actvn)) * scale;            % scale Activation to 0-1.

        % Now convert the activation into a firing rate using a binary or linear threshold activation functon
        % The threshold is selected in this case artificially by sorting
        % the array to determine where the threshold should be set to achieve the required sparseness
        % In the brain, inhibitory neurons are important in setting the proportion of excitatory neurons that are active
        tmp = sort(Actvn, 'descend');
        el = floor(length(Rate) * OutputSparseness);
        if el < 1
            el = 1;
        end
        Threshold = tmp(el);
        Rate = Actvn;
        % Threshold
        % Actvn
        for neuron = 1 : length(Rate)
            if Rate(neuron) >= Threshold % threshold binary or linear activation function
                Rate(neuron) = 1; % comment out this line for a threshold linear activation function
            else
                Rate(neuron) = 0;
            end
        end
        % now change the weights in the synaptic matrix. All weights will be >= 0
        for neuron = 1 : N
            for syn = 1 : nSyn % do the synapses for the compet external inputs
                preSynRate = TrainPatts(syn, patt);
                weight_change = Learnrate * Rate(neuron) * preSynRate;
                % weight_change = Learnrate * Rate(neuron) * (preSynRate - MeanPreSynapticRate); % OR try an associative rule with heterosynaptic LTD
                if weight_change > 0
                    % disp(['wtchange= ',num2str(weight_change), ' preSynrate=', num2str(preSynRate), ' Rate=', num2str(Rate(neuron)), ' neuron=', num2str(neuron), ' synapse=', num2str(syn)]); 
                end
                SynMat(syn + nSyn*2, neuron) = SynMat(syn + nSyn*2, neuron) + weight_change;
            end
            for syn = 1 : nSyn  % do the synapses for the recurrent attractor inputs
                preSynRate = Rate(syn);
                weight_change = LearnrateRC * Rate(neuron) * preSynRate;
                % weight_change = Learnrate * Rate(neuron) * (preSynRate - MeanPreSynapticRate); % OR try an associative rule with heterosynaptic LTD
                if weight_change > 0
                    % disp(['wtchange= ',num2str(weight_change), ' preSynrate=', num2str(preSynRate), ' Rate=', num2str(Rate(neuron)), ' neuron=', num2str(neuron), ' synapse=', num2str(syn)]); 
                end
                SynMat(syn + nSyn, neuron) = SynMat(syn + nSyn, neuron) + weight_change;
            end
            for syn = 1 : nSyn  % do the synapses for the BP inputs
                preSynRate = BPpatts(syn, patt);
                weight_change = LearnrateBP * Rate(neuron) * preSynRate;
                % weight_change = Learnrate * Rate(neuron) * (preSynRate - MeanPreSynapticRate); % OR try an associative rule with heterosynaptic LTD
                if weight_change > 0
                    % disp(['wtchange= ',num2str(weight_change), ' preSynrate=', num2str(preSynRate), ' Rate=', num2str(Rate(neuron)), ' neuron=', num2str(neuron), ' synapse=', num2str(syn)]); 
                end
                SynMat(syn, neuron) = SynMat(syn, neuron) + weight_change;
            end
            SynMat(:, neuron) = NormVeclen(SynMat(:, neuron)); % normalize the neuron's synaptic weight vector    
        end
        
        if display == 1
            figure(100);
            %subplot(2,2,2);
            positionVector1 = [0.2, 0.25, 0.7, 0.7]; % left, bottom, width, height. SynMat
            subplot('Position', positionVector1);
            imagesc(SynMat); title('synaptic matrix');
            colormap(gray);
            %subplot(2,2,4);
            positionVector2 = [0.2, 0.125, 0.7, 0.03]; % left, bottom, width, height. Activation
            subplot('Position', positionVector2);
            clims = [0 1];
            imagesc(Actvn); title('Activation'); % the internal recall term
            % disp(['Actvn max = ', num2str(max(Actvn)), ' min = ', num2str(min(Actvn))]);
            set (gca, 'YtickLabel', '');
            positionVector3 = [0.2, 0.0, 0.7, 0.03]; % left, bottom, width, height. Rate
            subplot('Position', positionVector3);
            clims = [0 1];
            %imagesc(Rate, clims); title('Postsynaptic Rate'); % the internal recall term
            imagesc(Rate); title('Postsynaptic Rate'); % the internal recall term
            % disp(['Rate max = ', num2str(max(Rate)), ' min = ', num2str(min(Rate))]);
            set (gca, 'YtickLabel', '');
            %subplot(2,2,1);
            positionVector4 = [0.1, 0.25, 0.02, 0.7]; % left, bottom, width, height. PreSynaptic Input
            subplot('Position', positionVector4);
            imagesc(PreSynInput);
            title('Presynaptic input');
            set (gca, 'XtickLabel', '');
            drawnow;
%             patt
%             pause;
        end
    end
end
%SynMat

% Now test it with the forward patterns applied once, and letting the recurrent collaterals continue the firing
disp('Touch Enter to start Testing for short-term memory after a training pattern is applied once');
pause
RecalledRates = zeros(N, nPatts); % the firing rate output vectors have N elements
ZInputVector = zeros(nSyn, 1); % zero presynaptic input vector
sparsenessSum = 0;
for patt = 1 : nPatts
    for RecEpochs = 1 : NRecEpochs
        % calculate activation
        for neuron = 1 : N
            PreSynInput = ZInputVector;                % there is no BP input yet
            if RecEpochs == 1
                PreSynInput = [PreSynInput; ZInputVector]; % use just the external forward input on recall epoch 1
                PreSynInput = [PreSynInput; TrainPatts(:, patt)]; % the forward input retrieval cue
            else
                PreSynInput = [PreSynInput; Rate(:) * ScaleFactAttToComp]; % do short-term memory with the attractor synapses
                PreSynInput = [PreSynInput; ZInputVector]; % there is forward input while we do short-term memory
            end
            Actvn(neuron) = dot(PreSynInput,SynMat(:, neuron)); % dot product of input pattern and synaptic weight vector for one neuron
        end
        scale = 1.0 / (max(Actvn) - min(Actvn));
        Actvn = (Actvn - min(Actvn)) * scale;            % scale Activation to 0-1.

        % Now convert the activation into a firing rate using a binary or linear threshold activation functon
        % The threshold is selected in this case artificially by sorting
        % the array to determine where the threshold should be set to achieve the required sparseness
        % In the brain, inhibitory neurons are important in setting the proportion of excitatory neurons that are active
        tmp = sort(Actvn, 'descend');
        el = floor(length(Rate) * OutputSparseness);
        if el < 1
            el = 1;
        end
        Threshold = tmp(el);
        Rate = Actvn;
        % Threshold
        % Actvn
        for neuron = 1 : N
            if Rate(neuron) >= Threshold % threshold binary or linear activation function
                Rate(neuron) = 1; % comment out this line for a threshold linear acivation function
            else
                Rate(neuron) = 0;
            end
        end
        RecalledRates(:, patt) = Rate; % save the recalled rate for later analysis
        if display == 1
            figure(101);
            %subplot(2,2,2);
            positionVector1 = [0.2, 0.25, 0.7, 0.7]; % left, bottom, width, height. SynMat
            subplot('Position', positionVector1);
            imagesc(SynMat); title('synaptic matrix');
            colormap(gray);
            %subplot(2,2,4);
            positionVector2 = [0.2, 0.125, 0.7, 0.03]; % left, bottom, width, height. Activation
            subplot('Position', positionVector2);
            clims = [0 1];
            imagesc(Actvn); title('Activation'); % the internal recall term
            % disp(['Actvn max = ', num2str(max(Actvn)), ' min = ', num2str(min(Actvn))]);
            set (gca, 'YtickLabel', '');
            positionVector3 = [0.2, 0.0, 0.7, 0.03]; % left, bottom, width, height. Rate
            subplot('Position', positionVector3);
            clims = [0 1];
            %imagesc(Rate, clims); title('Postsynaptic Rate'); % the internal recall term
            imagesc(Rate); title('Postsynaptic Rate'); % the internal recall term
            % disp(['Rate max = ', num2str(max(Rate)), ' min = ', num2str(min(Rate))]);
            set (gca, 'YtickLabel', '');
            %subplot(2,2,1);
            positionVector4 = [0.1, 0.25, 0.02, 0.7]; % left, bottom, width, height. PreSynaptic Input
            subplot('Position', positionVector4);
            imagesc(PreSynInput);
            title('Presynaptic input');
            set (gca, 'XtickLabel', '');
            drawnow;
            sparse = spars(Rate);
            sparsenessSum = sparsenessSum + sparse;
            %pause;
        end
        fprintf('Pattern=%2d  RecEpoch = %2d  Neurons=', patt, RecEpochs);
        I = find(Rate>0.5);
        for j = 1 : length(I)
            fprintf('%2d  ', I(j));
        end
        fprintf('\n');
        %if RecEpochs == 1; pause; end 
        %pause(0.2)
    end
end

MeanSparseness = sparsenessSum / nPatts;
%pause 

% show correlation matrix between the different training patterns for comparison
PattCorMat = zeros(nPatts);
for i = 1 : nPatts
    for j = 1 : nPatts
        R = corr(TrainPatts(:, i), TrainPatts(:, j));
        PattCorMat(j,i) = R;
    end
end
fig = fig + 1;
figure(fig)
imagesc(PattCorMat);
colormap(gray(256));
c = colorbar;
c.Label.String = 'Correlation';
title([' Correlation between input patterns ', '  min=', num2str(min(min(PattCorMat))), '  max=', num2str(max(max(PattCorMat)))]);
xlabel('Pattern number'); ylabel('Pattern number');
%[Pattcoeff,Pattlatent,Pattexplained] = pcacov(PattCorMat);
% and get the mean abs value of the off-diagonal terms, which are the
% correlations between the different patterns
numvals = 0;
PattsumCor = 0;
for col = 1 : nPatts
    for row = 1 : nPatts
        if row ~= col && PattCorMat(row, col) < 0.99
            PattsumCor = PattsumCor + abs(PattCorMat(row, col));
            numvals = numvals + 1;
        end
    end
end
disp(['Average of absolute values of correlation < 1.0 between different input patterns=', num2str(PattsumCor / numvals)]); 

% show correlation matrix between the output rates for the different training patterns
CorMat = zeros(nPatts);
for i = 1 : nPatts
    for j = 1 : nPatts
        R = corr(RecalledRates(:, i), RecalledRates(:, j));
        CorMat(j,i) = R;
    end
end
fig = fig + 1;
figure(fig)
x = linspace(-1, 1, 21);
imagesc(CorMat);
colormap(gray(256));
c = colorbar;
c.Label.String = 'Correlation';
%title([' Correlation between Output Rates ', '  min=', num2str(min(min(CorMat))), '  max=', num2str(max(max(CorMat)))]);
title([' Correlation between Output Rates using Forward Inputs']);
xlabel('Pattern number'); ylabel('Pattern number');

% and get the mean abs value of the off-diagonal  < 1 terms, which are the
% correlations between the rates output for the different patterns
numvals = 0;
sumCor = 0;
for col = 1 : nPatts
    for row = 1 : nPatts
        if row ~= col && CorMat(row, col) < 0.99
            sumCor = sumCor + abs(CorMat(row, col));
            numvals = numvals + 1;
        end
    end
end
disp(['Average of absolute values of correlation < 1.0 between output Rates to different patterns=', num2str(sumCor / numvals)]); 

% ------------------------------------------------------------------------------------ %
% Now test it with the back projections applied once, and letting the recurrent collaterals continue the firing
disp('Touch Enter to start Testing for short-term memory after a BackProjection pattern is applied once');
pause
RecalledRates = zeros(N, nPatts); % the firing rate output vectors have N elements
sparsenessSum = 0;
for patt = 1 : nPatts
    for RecEpochs = 1 : NRecEpochs
        % calculate activation
        for neuron = 1 : N
            if RecEpochs == 1
                PreSynInput = BPpatts(:, patt) * ScaleFactBP; % the BP input on epoch 1
                PreSynInput = [PreSynInput; ZInputVector; ZInputVector]; % use just the backprojection input on recall epoch 1
            else
                PreSynInput = [ZInputVector; Rate(:) * ScaleFactAttToComp]; % no recall cue, just attractor synapses
                PreSynInput = [PreSynInput; ZInputVector]; % there is forward input now, just recurrent collateral input
            end
            Actvn(neuron) = dot(PreSynInput,SynMat(:, neuron)); % dot product of input pattern and synaptic weight vector for one neuron
        end
        scale = 1.0 / (max(Actvn) - min(Actvn));
        Actvn = (Actvn - min(Actvn)) * scale;            % scale Activation to 0-1.

        % Now convert the activation into a firing rate using a binary or linear threshold activation functon
        % The threshold is selected in this case artificially by sorting
        % the array to determine where the threshold should be set to achieve the required sparseness
        % In the brain, inhibitory neurons are important in setting the proportion of excitatory neurons that are active
        tmp = sort(Actvn, 'descend');
        el = floor(length(Rate) * OutputSparseness);
        if el < 1
            el = 1;
        end
        Threshold = tmp(el);
        Rate = Actvn;
        % Threshold
        % Actvn
        for neuron = 1 : N
            if Rate(neuron) >= Threshold % threshold binary or linear activation function
                Rate(neuron) = 1; % comment out this line for a threshold linear activation function
            else
                Rate(neuron) = 0;
            end
        end
        RecalledRates(:, patt) = Rate; % save the recalled rate for later analysis
        if display == 1
            figure(101);
            %subplot(2,2,2);
            positionVector1 = [0.2, 0.25, 0.7, 0.7]; % left, bottom, width, height. SynMat
            subplot('Position', positionVector1);
            imagesc(SynMat); title('synaptic matrix');
            colormap(gray);
            %subplot(2,2,4);
            positionVector2 = [0.2, 0.125, 0.7, 0.03]; % left, bottom, width, height. Activation
            subplot('Position', positionVector2);
            clims = [0 1];
            imagesc(Actvn); title('Activation'); % the internal recall term
            % disp(['Actvn max = ', num2str(max(Actvn)), ' min = ', num2str(min(Actvn))]);
            set (gca, 'YtickLabel', '');
            positionVector3 = [0.2, 0.0, 0.7, 0.03]; % left, bottom, width, height. Rate
            subplot('Position', positionVector3);
            clims = [0 1];
            %imagesc(Rate, clims); title('Postsynaptic Rate'); % the internal recall term
            imagesc(Rate); title('Postsynaptic Rate'); % the internal recall term
            % disp(['Rate max = ', num2str(max(Rate)), ' min = ', num2str(min(Rate))]);
            set (gca, 'YtickLabel', '');
            %subplot(2,2,1);
            positionVector4 = [0.1, 0.25, 0.02, 0.7]; % left, bottom, width, height. PreSynaptic Input
            subplot('Position', positionVector4);
            imagesc(PreSynInput);
            title('Presynaptic input');
            set (gca, 'XtickLabel', '');
            drawnow;
            sparse = spars(Rate);
            sparsenessSum = sparsenessSum + sparse;
            %pause;
        end
        fprintf('Pattern=%2d  RecEpoch = %2d  Neurons=', patt, RecEpochs);
        I = find(Rate>0.5);
        for j = 1 : length(I)
            fprintf('%2d  ', I(j));
        end
        fprintf('\n');
        %if RecEpochs == 1; pause; end 
        %pause(0.2)
    end
end

MeanSparseness = sparsenessSum / nPatts;
%pause 


% show correlation matrix between the output rates for the different training patterns
CorMat = zeros(nPatts);
for i = 1 : nPatts
    for j = 1 : nPatts
        R = corr(RecalledRates(:, i), RecalledRates(:, j));
        CorMat(j,i) = R;
    end
end
fig = fig + 1;
figure(fig)
x = linspace(-1, 1, 21);
imagesc(CorMat);
colormap(gray(256));
c = colorbar;
c.Label.String = 'Correlation';
title([' Correlation between Output Rates using Backprojections']);
xlabel('Pattern number'); ylabel('Pattern number');

% and get the mean abs value of the off-diagonal terms that are < 1, which are the
% correlations between the rates output for the different patterns
numvals = 0;
sumCor = 0;
for col = 1 : nPatts
    for row = 1 : nPatts
        if row ~= col && CorMat(row, col) < 0.99
            sumCor = sumCor + abs(CorMat(row, col));
            numvals = numvals + 1;
        end 
    end
end
disp(['Average of absolute values of correlations < 1.0 between output Rates to different patterns=', num2str(sumCor / numvals)]); 
%[coeff,latent,explained] = pcacov(CorMat);


 

